@@ -42,9 +42,18 @@ import getpass
 import email
 import poplib
 import imaplib
-import sets
+
+try:
+    # do we have a recent pykerberos?
+    HAVE_KERBEROS_GSS = False
+    import kerberos
+    if 'authGSSClientWrap' in dir(kerberos):
+        HAVE_KERBEROS_GSS = True
+except ImportError:
+    pass
 
 from getmailcore.exceptions import *
+from getmailcore.compatibility import *
 from getmailcore.constants import *
 from getmailcore.message import *
 from getmailcore.utilities import *
@@ -61,6 +70,16 @@ NOT_ENVELOPE_RECIPIENT_HEADERS = (
     'resent-bcc'
 )
 
+# How long a vanished message is kept in the oldmail state file for IMAP
+# retrievers before we figure it's gone for good.  This is to allow users
+# to only occasionally retrieve mail from certain IMAP folders without
+# losing their oldmail state for that folder.  This is in seconds, so it's
+# 30 days.
+VANISHED_AGE = (60 * 60 * 24 * 30)
+
+# Kerberos authentication state constants
+(GSS_STATE_STEP, GSS_STATE_WRAP) = (0, 1)
+
 #
 # Mix-in classes
 #
@@ -331,12 +350,19 @@ class RetrieverSkeleton(ConfigurableBase):
         '''Read contents of oldmail file.'''
         self.log.trace()
         try:
-            for (msgid, timestamp) in [
-                line.strip().split('\0', 1) for line in open(
-                    self.oldmail_filename, 'rb'
-                ) if (line.strip()!='' and '\0' in line)
-            ]:
-                self.oldmail[msgid] = int(timestamp)
+            for line in open(self.oldmail_filename, 'rb'):
+                line = line.strip()
+                if not line or not '\0' in line:
+                    # malformed
+                    continue
+                try:
+                    (msgid, timestamp) = line.split('\0', 1)
+                    self.oldmail[msgid] = int(timestamp)
+                except ValueError:
+                    # malformed
+                    self.log.info('skipped malformed line "%r" for %s'
+                                  % (line, self)
+                                  + os.linesep)
             self.log.moreinfo('read %i uids for %s' % (len(self.oldmail), self)
                               + os.linesep)
         except IOError:
@@ -351,9 +377,9 @@ class RetrieverSkeleton(ConfigurableBase):
         wrote = 0
         try:
             f = updatefile(self.oldmail_filename)
-            msgids = sets.ImmutableSet(
+            msgids = frozenset(
                 self.__delivered.keys()
-            ).union(sets.ImmutableSet(self.oldmail.keys()))
+            ).union(frozenset(self.oldmail.keys()))
             for msgid in msgids:
                 self.log.debug('msgid %s ...' % msgid)
                 if forget_deleted and msgid in self.deleted:
@@ -509,7 +535,10 @@ class POP3RetrieverBase(RetrieverSkeleton):
             msg = Message(fromlines=lines)
             return msg
         except poplib.error_proto, o:
-            raise getmailOperationError('POP error (%s)' % o)
+            raise getmailRetrievalError(
+                'failed to retrieve msgid %s; server said %s' 
+                % (msgid, o)
+            )
 
     def _getheaderbyid(self, msgid):
         self.log.trace()
@@ -637,6 +666,41 @@ class IMAPRetrieverBase(RetrieverSkeleton):
         self.log.trace()
         self.mailbox = None
         self.uidvalidity = None
+        self.gss_step = 0
+        self.gss_vc = None
+        self.gssapi = False
+
+    def checkconf(self):
+        RetrieverSkeleton.checkconf(self)
+        if self.conf['use_kerberos'] and not HAVE_KERBEROS_GSS:
+            raise getmailConfigurationError(
+                'cannot use kerberos authentication; Python kerberos support '
+                'not installed or does not support GSS'
+            )
+
+    def gssauth(self, response):
+        if not HAVE_KERBEROS_GSS:
+            # shouldn't get here
+            raise ValueError('kerberos GSS support not available')
+        data = ''.join(str(response).encode('base64').splitlines())
+        if self.gss_step == GSS_STATE_STEP:
+            if not self.gss_vc:
+                (rc, self.gss_vc) = kerberos.authGSSClientInit(
+                    'imap@%s' % self.conf['server']
+                )
+                response = kerberos.authGSSClientResponse(self.gss_vc)
+            rc = kerberos.authGSSClientStep(self.gss_vc, data)
+            if rc != kerberos.AUTH_GSS_CONTINUE:
+               self.gss_step = GSS_STATE_WRAP
+        elif self.gss_step == GSS_STATE_WRAP:
+            rc = kerberos.authGSSClientUnwrap(self.gss_vc, data)
+            response = kerberos.authGSSClientResponse(self.gss_vc)
+            rc = kerberos.authGSSClientWrap(self.gss_vc, response,
+                                            self.conf['username'])
+        response = kerberos.authGSSClientResponse(self.gss_vc)
+        if not response:
+            response = ''
+        return response.decode('base64')
 
     def __del__(self):
         RetrieverSkeleton.__del__(self)
@@ -794,7 +858,15 @@ class IMAPRetrieverBase(RetrieverSkeleton):
             # Retrieve message
             self.log.debug('retrieving body for message "%s"' % uid
                            + os.linesep)
-            response = self._parse_imapuidcmdresponse('FETCH', uid, part)
+            try:
+                response = self._parse_imapuidcmdresponse('FETCH', uid, part)
+            except (imaplib.IMAP4.error, getmailOperationError), o:
+                # server gave a negative/NO response, most likely.  Bad server,
+                # no doughnut.
+                raise getmailRetrievalError(
+                    'failed to retrieve msgid %s; server said %s' 
+                    % (msgid, o)
+                )
             # Response is really ugly:
             #
             # [
@@ -805,7 +877,16 @@ class IMAPRetrieverBase(RetrieverSkeleton):
             #   ')',
             #   <maybe more>
             # ]
-            msg = Message(fromstring=response[0][1])
+            
+            # MSExchange is broken -- if a message is badly formatted enough
+            # (virus, spam, trojan), it can completely fail to return the
+            # message when requested.
+            try:
+                msg = Message(fromstring=response[0][1])
+            except TypeError, o:
+                # response[0] is None instead of a message tuple
+                raise getmailRetrievalError('failed to retrieve msgid %s' 
+                                            % msgid)
             return msg
 
         except imaplib.IMAP4.error, o:
@@ -822,15 +903,19 @@ class IMAPRetrieverBase(RetrieverSkeleton):
     def initialize(self):
         self.log.trace()
         # Handle password
-        if self.conf.get('password', None) is None:
-            self.conf['password'] = getpass.getpass('Enter password for %s:  '
-                                                    % self)
+        if (self.conf.get('password', None) is None
+                and not (HAVE_KERBEROS_GSS and self.conf['use_kerberos'])):
+            self.conf['password'] = getpass.getpass(
+                'Enter password for %s:  ' % self
+            )
         RetrieverSkeleton.initialize(self)
         try:
             self.log.trace('trying self._connect()' + os.linesep)
             self._connect()
             self.log.trace('logging in' + os.linesep)
-            if self.conf['use_cram_md5']:
+            if self.conf['use_kerberos'] and HAVE_KERBEROS_GSS:
+                self.conn.authenticate('GSSAPI', self.gssauth)
+            elif self.conf['use_cram_md5']:
                 self._parse_imapcmdresponse(
                     'login_cram_md5', self.conf['username'],
                     self.conf['password']
@@ -843,10 +928,15 @@ class IMAPRetrieverBase(RetrieverSkeleton):
             self.log.debug('msgids: %s'
                            % sorted(self.msgnum_by_msgid.keys()) + os.linesep)
             self.log.debug('msgsizes: %s' % self.msgsizes + os.linesep)
-            # Remove messages from state file that are no longer in mailbox
+            # Remove messages from state file that are no longer in mailbox,
+            # but only if the timestamp for them are old (30 days for now).
+            # This is because IMAP users can have one state file but multiple
+            # IMAP folders in different configuration rc files.
             for msgid in self.oldmail.keys():
-                if not self.msgsizes.has_key(msgid):
-                    self.log.debug('removing vanished message id %s' % msgid
+                timestamp = self.oldmail[msgid]
+                age = self.timestamp - timestamp
+                if not self.msgsizes.has_key(msgid) and age > VANISHED_AGE:
+                    self.log.debug('removing vanished old message id %s' % msgid
                                    + os.linesep)
                     del self.oldmail[msgid]
         except imaplib.IMAP4.error, o:
